import java.util. *;
import java.awt. *;
import java.io. *;
import java.net. *;
import java.applet.Applet;

/** Execute a splay tree operation on a binary search tree. The
 ** tree implementation does not currently use sentinals. Null 
 ** Node references are treated as if they refer to black Nodes. */

class RedBlackTreeOperation 
extends DictionaryOperation implements Runnable {

  /** Initialize with a given Action, snapshot of the current graph
   ** (SimpleGraph), the tree's panel and an IconLibrary */

    RedBlackTreePanel rbtreePanel;

    RedBlackTreeOperation(Action a, Tree t, SimpleGraph sg, 
			  RedBlackTreePanel p, IconLibrary ih) {
	super(a, t, sg, (GraphPanel) p, ih);
	rbtreePanel = p;
    }
			  
  /** returns true when given a non-null reference to a node flagged red */ 
  boolean isRed(Node n) { 
    return (n != null && n.redFlag);
  }
  /** returns true if the given Node referece is black or null */
  boolean isBlack(Node n) {
    return (n == null || !n.redFlag);
  }
  /** recolor red and set graphics on non-null references only */
  void colorRed(Node n) {
    if (n != null) {
      n.fixGraphics(IconLibrary.RED);
      n.redFlag = true;
    }
  }
  /** recolor black and set graphics on non-null references only */
  void colorBlack(Node n) {
    if (n != null) {
      n.fixGraphics(IconLibrary.BLACK);
      n.redFlag = false;
    }
  }

  private boolean simpleRecoloring;

  /** execute the RedBlackOperation after binary search tree operations are
   ** completed */
  public void prologue(int op, Node n) 
  throws InterruptedException {

      if (n != null) 
	  switch (op) {

	  case INSERT:
	      colorRed(n);
	      break;
	      
	  case DELETE:
	    break;
	      
	  default :
	      break;
	  }
  }

  /** execute the RedBlack after binary search tree operations are
     ** completed */
  public void epilogue(int op, Node n) 
  throws InterruptedException {

      if (n != null) 
	  switch (op) {

	  case INSERT:
	      recolor(n);
	      waitToSettle(n);
	      break;
	      
	  case DELETE:
	    break;
	      
	  default:
	      break;
	  }

      rbtreePanel.spots(null);
      panel.caption("Red-black properties satisfied.");
  }

  void pauseBetweenSteps() { 
      tree.clear();
      try longPause(1000);
      catch (InterruptedException e) {}
  }

  /** slide the tree's finder (little ball) between a pair of nodes
   ** but only if both nodes are non-null */
  void slide(Node n, Node m) {
    if (n != null && m != null) 
      tree.finder.slide(n,m);
  }

  /** recoloring after insertion */
  void recolor (Node n) 
  throws InterruptedException {

    Node p; // parent of n
    Node g; // grandparent of n
    Node u; // uncle of n (i.e. p's sibling)

    panel.title("Recolor:");
    tree.clear();

    rbtreePanel.spots(n); // set spotlight on n

    boolean simpleCase = false;

    if (n == null || isBlack(n)) {
        panel.caption("Root is already black.");
	simpleCase = true;
    }
    else if (n.isRoot()) { // must be red
        panel.caption("Case 1: Recolor the root to black. Stop.");
	pauseBetweenSteps();
        colorBlack(n);
	simpleCase = true;
    }
    else if (isBlack(n.parent)) {
        panel.caption("Case 2: This red node has a black parent. Stop.");
	n.found(true);
	slide(n,n.parent);
	colorRed(n);
	simpleCase = true;
    }

    if (simpleCase) {
	pauseBetweenSteps();
	tree.clear();
	rbtreePanel.spots(null);
	return;
    }

    p = n.parent;
    g = p.parent;
    u = p.sibling();

    if (isRed(u)) {

        panel.caption("Case 3: Split, and repeat from the grandparent"); 

	n.found(true);
	slide(n,p);
	p.found(true);
	slide(p,g);
	if (u != null) 
	    u.found(true);

	if (rbtreePanel.layout.type == Layout.RB234) {
	    colorRed(g);
	    colorBlack(u);
	    colorBlack(p);
	    tree.traverse(snapshot);
	}
	else {
	    tree.finder.fixGraphics(IconLibrary.GRAY);
	    colorRed(g);
	    (tree.finder.duplicate(g,u)).fixGraphics(IconLibrary.GRAY);
	    slide(g,p);
	    colorBlack(u);
	    colorBlack(p);
	    tree.finder.unfixGraphics();
	}
	tree.finder.moveTo(g);
	pauseBetweenSteps();
	colorRed(n);
	rbtreePanel.spots(g);
	recolor(g);
    }
    else {     /* uncle u is black */

	if (n.parent.isLeft() != n.isLeft()) { 
	    panel.caption("Case 4b: Terminate with a double rotation.");
	    rotate(snapshot, n);
	    if (rbtreePanel.layout.type == Layout.RB234)
		tree.traverse(snapshot);
	    waitToSettle();

	    colorRed(n);
	    slide(n, p);
	    n.found(true);
	    p.found(true);
	    if (u != null) 
		u.found(true);

	    Node tmp = p;
	    p = n;
	    n = tmp;

	    rbtreePanel.spots(n);
	    pauseBetweenSteps();
	}
	else 
	    panel.caption("Case 4a: Terminate with a single rotation.");

	p.found(false);
	slide(n, p);
	n.found(true);
	colorRed(g);
	colorBlack(p);
	if (rbtreePanel.layout.type != Layout.RB234) {
	    p.exchange(g);
//	    waitToSettle();
	}
        rotate(snapshot, p);
	colorRed(n);
	if (rbtreePanel.layout.type == Layout.RB234)
	    tree.traverse(snapshot);
	briefPause();
    }
    tree.clear();
    rbtreePanel.spots(null);
    waitToSettle();
    pauseBetweenSteps();
  }
}

/** */
class RedBlackTreePanel extends GraphPanel implements Runnable {

  Node n, nParent, nGrandParent, nUncle;

  RedBlackTreePanel (RedBlackTree st, Layout l, Dimension dim, IconLibrary im) {
    super((Graph) st, l, dim, im);
    n = nParent = nUncle = nGrandParent = null;
  }

  void spots(Node node) {
      follow = node;
  }

  public void launchAction(int action) {

    if (rotating || pick == null) return;

    Action a = new Action();
    a.op = action;

    switch(action) {

	  case DictionaryOperation.INSERT : 
	    	a.node = insertAndPlace(getInsertValue(pick));
	    	a.node.x = pick.x;
	    	a.node.y = pick.y;
	       	break;

	  case DictionaryOperation.DELETE :
	  case DictionaryOperation.LOCATE :
	    	a.node = pick;
	    	break;

	  default :
	    	break;
    }

    nextAction(a);
    unpick();
  }

  /** */
  public synchronized void startOperation() {
    rotating = true;
    super.startOperation();
  }

  /** */
  public synchronized void endOperation() {
    rotating = false;
    super.endOperation();
  }

  protected void animate (Action a) {

    if (layout.demoMode) {
	follow = a.node;
    }
    RedBlackTreeOperation RedBlack; 
    RedBlack = new RedBlackTreeOperation(a,tree,snapshot,this,images);
    Thread t = new Thread(RedBlack);
    tree.continuation(t);
    t.start();
  }
}

public class RedBlackTree extends Graph {

  RedBlackTreePanel rbtPanel;

  protected void launchAction (Object arg) {
     if ("Insert".equals (arg)) 
        rbtPanel.launchAction(DictionaryOperation.INSERT);
     else if ("Delete".equals (arg)) {
     //	System.out.println("Sorry. Deletion not yet implemented for rbtree."); 
	rbtPanel.launchAction(DictionaryOperation.DELETE);
     }
     else if ("Locate".equals (arg))
	rbtPanel.launchAction(DictionaryOperation.LOCATE);
  }

  public void init() {
      initPrelude();

      Dimension d = size();
      Dimension pDim = new Dimension(d.width, d.height-100);

      rbtPanel = new RedBlackTreePanel (this, layout, pDim, images);
      panel = rbtPanel;

      add ("Center", panel);
      showChoices(true);
      showButtons(false);

      layout.demoMode = true; // can't use colors to follow nodes here
  }
  public void start() {
      rbtPanel.initialTree = edges;
      rbtPanel.start();
  }
  public void stop() {
      rbtPanel.stop();
  }
}

